@@ -1,17 +1,19 @@
 package Class::InsideOut;
+use strict;
 
-$VERSION     = "1.03";
+use vars qw/$VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS/;
+
+$VERSION     = '1.09';
 @ISA         = qw ( Exporter );
 @EXPORT      = qw ( ); # nothing by default
-@EXPORT_OK   = qw ( new id options private property public register );
+@EXPORT_OK   = qw ( new id options private property public readonly register );
 %EXPORT_TAGS = (
-    "std"       => [ qw( id private public register ) ],
+    "std"       => [ qw( id private public readonly register ) ],
     "new"       => [ qw( new ) ],
     "all"       => [ @EXPORT_OK ],
     "singleton" => [], # just a flag for import()
 );
 
-use strict;
 use Carp;
 use Exporter;
 use Class::ISA;
@@ -22,8 +24,8 @@ BEGIN {
     eval { Scalar::Util->import( "weaken" ) };
     if ( $@ =~ /\AWeak references/ ) {
         warn "Scalar::Util::weaken unavailable: "
-           . "Class::InsideOut will not be thread-safe\n";
-        *weaken = sub { shift };
+           . "Class::InsideOut will not be thread-safe and will leak memory\n";
+        *weaken = sub { return @_ };
     }
 }
 
@@ -67,10 +69,10 @@ sub import {
     my $caller = caller;
     *{ "$caller\::DESTROY" } = _gen_DESTROY( $caller );
     # check for ":singleton" and do export attach instead of thaw
-    # make ":singleton" an empty tag to Exporter doesn't choke on it
     if ( grep { $_ eq ":singleton" } @_ ) {
         *{ "$caller\::STORABLE_freeze" } = _gen_STORABLE_freeze( $caller, 1 );
         *{ "$caller\::STORABLE_attach" } = _gen_STORABLE_attach( $caller );
+        @_ = grep { $_ ne ':singleton' } @_; # strip it back out
     }
     else {
         *{ "$caller\::STORABLE_freeze" } = _gen_STORABLE_freeze( $caller, 0 );
@@ -97,7 +99,8 @@ sub new {
     
     # initialization
     croak "Arguments to new must be a hash or hash reference"
-        if ( @_ == 1 && ref($_[0]) && reftype($_[0]) ne 'HASH' ) || ( @_ % 2 );
+        if ( @_ == 1 && ! ( ref $_[0] && reftype($_[0]) eq 'HASH' ) ) 
+        || ( @_ > 1 && @_ % 2 );
      
     my %args = (@_ == 1) ? %{$_[0]} : @_;
 
@@ -114,25 +117,36 @@ sub new {
     return $self;
 }
 
-sub private($\%;$) {
+sub private($\%;$) { ## no critic -- prototype
     &_check_property;
     $_[2] ||= {};
     $_[2] = { %{$_[2]}, privacy => 'private' };
     goto &_install_property;
 }
 
-sub property($\%;$) {
+sub property($\%;$) { ## no critic -- prototype
     &_check_property;
     goto &_install_property;
 }
 
-sub public($\%;$) {
+sub public($\%;$) { ## no critic -- prototype
     &_check_property;
     $_[2] ||= {};
     $_[2] = { %{$_[2]}, privacy => 'public' };
     goto &_install_property;
 }
 
+sub readonly($\%;$) { ## no critic -- prototype
+    &_check_property;
+    $_[2] ||= {};
+    $_[2] = { 
+        %{$_[2]}, 
+        privacy => 'public',
+        set_hook => sub { die "is read-only\n" }
+    };
+    goto &_install_property;
+}
+
 sub register {
     my ($obj);
     if    ( @_ == 0 ) {
@@ -163,7 +177,6 @@ sub register {
     return $obj;
 }
 
-
 #--------------------------------------------------------------------------#
 # private functions for implementation
 #--------------------------------------------------------------------------#
@@ -194,7 +207,7 @@ sub CLONE {
 
         # update the registry to the new, cloned object
         weaken ( $OBJECT_REGISTRY{ $new_id } = $object );
-        delete $OBJECT_REGISTRY{ $old_id };
+        _deregister( $old_id );
     }
 }
 
@@ -220,7 +233,7 @@ sub _check_property {
     my ($label, $hash, $opt) = @_;
     local $Carp::CarpLevel = $Carp::CarpLevel + 1;
     croak "Invalid property name '$label': must be a perl identifier"
-        if $label !~ /\A[a-z_]\w*\z/;
+        if $label !~ /\A[a-z_]\w*\z/i;
     croak "Duplicate property name '$label'"
         if grep { $_ eq $label } keys %{ $PROP_DATA_FOR{ caller(1) } }; 
     _check_options( $opt ) if defined $opt;
@@ -233,6 +246,14 @@ sub _class_tree {
     return @{ $CLASS_ISA{ $class } };
 }
 
+# take either object or object id
+sub _deregister {
+    my ($arg) = @_;
+    my $obj_id = ref $arg ? refaddr $arg : $arg;
+    delete $OBJECT_REGISTRY{ $obj_id };
+    return;
+}
+
 # turn object into hash -- see _revert()
 sub _evert {
     my ( $obj ) = @_;
@@ -336,7 +357,7 @@ sub _gen_DESTROY {
         # it should only call SUPER::DEMOLISH if need be; still,
         # rest of the destructor doesn't need the registry, so early deletion
         # by a subclass should be safe
-        delete $OBJECT_REGISTRY{ $obj_id };
+        _deregister( $obj );
 
         return;
     };
@@ -364,8 +385,9 @@ sub _gen_STORABLE_attach {
             return $class->new();
         }
         else {
-            die "Error attaching to $class:\n" .
+            warn "Error attaching to $class:\n" .
                   "Couldn't find STORABLE_attach_hook() or new() in $class\n";
+            return;
         }
     };
 }
@@ -538,10 +560,11 @@ This documentation refers to version %%VERSION%%
 
  package My::Class;
  
- use Class::InsideOut qw( public private register id );
+ use Class::InsideOut qw( public readonly private register id );
 
- public     name => my %name;       # accessor: name()
- private    age  => my %age;        # no accessor
+ public     name => my %name;    # accessor: name()
+ readonly   ssn  => my %ssn;     # read-only accessor: ssn()
+ private    age  => my %age;     # no accessor
  
  sub new { register( shift ) }
  
@@ -561,7 +584,7 @@ module aims for minimalism and robustness:
 * Supports any underlying object type including black-box inheritance
 * Does not leak memory on object destruction
 * Overloading-safe
-* Thread-safe for Perl 5.8 or better
+* Thread-safe for Perl 5.8.5 or better
 * {mod_perl} compatible
 * Makes no assumption about inheritance or initializer needs
 
@@ -598,7 +621,7 @@ Additional functions may be imported as usual by including them as arguments to
 As a shortcut, {Class::InsideOut} supports two tags for importing sets of
 functions:
 
-* {:std} provides {id}, {private}, {public} and {register}
+* {:std} provides {id}, {private}, {public}, {readonly} and {register}
 * {:all} imports all functions (including an optional constructor)
 
 *Note*: Automatic imports can be bypassed via {require} or by passing an empty
@@ -607,12 +630,13 @@ this is a good idea.
 
 == Object properties and accessors
 
-Object properties are declared with the {public} and {private}
+Object properties are declared with the {public}, {readonly} and {private}
 functions.  They must be passed a label and the lexical hash that will be
 used to store object properties:
 
- public name => my %name;
- private age => my %age;
+ public   name => my %name;
+ readonly ssn  => my %ssn;
+ private  age  => my %age;
 
 Properties for an object are accessed through an index into the lexical hash
 based on the memory address of the object.  This memory address ~must~ be
@@ -620,6 +644,7 @@ obtained via {Scalar::Util::refaddr}.  The alias {id} may be imported for
 brevity.
 
  $name{ refaddr $self } = "James";
+ $ssn { id      $self } = 123456789;
  $age { id      $self } = 32;
 
 *Tip*: since {refaddr} and {id} are function calls, it may be efficient to
@@ -635,6 +660,17 @@ the property.
  $person = My::Class->new;
  $person->name( "Larry" );
 
+Object properties declared with {readonly} will have a read-only accessor
+created.  The accessor will die if passed an argument to set the property
+value.  The property may be set directly in the hash from within the class
+package as usual.
+
+ # Inside the class
+ $ssn { id $person } = 987654321;
+ 
+ # Inside or outside the class
+ $person->ssn( 123456789 );      # dies
+
 Property accessors may also be hand-written by declaring the property
 {private} and writing whatever style of accessor is desired.  For example:
 
@@ -644,7 +680,17 @@ Property accessors may also be hand-written by declaring the property
 Hand-written accessors will be very slightly faster as generated accessors hold
 a reference to the property hash rather than accessing the property hash
 directly.
- 
+
+It is also possible to use a package hash instead of a lexical hash to store
+object properties:
+
+ public name => our %name;
+
+However, this makes private object data accessable outside the class and incurs
+a slight performance penalty when accessing the property hash directly; it is
+not recommended to do this unless you really need it for some specialized
+reason.
+
 == Object construction
 
 {Class::InsideOut} provides no default constructor method as there are many
@@ -676,7 +722,9 @@ case of blessing an anonymous scalar.
  register( $class ); # same as register( bless \(my $s), $class )
 
 As a convenience, {Class::InsideOut} provides an optional {new} constructor
-for simple objects.
+for simple objects.  This constructor automatically initializes the object
+from key/value pairs passed to the constructor for all keys matching the 
+name of a property (including otherwise "private" or "readonly" properties).
  
 A more advanced technique for object construction uses another object, usually
 a superclass object, as the object reference.  See "black-box inheritance" in
@@ -793,6 +841,16 @@ default options that have been set.
 This is an alias to {property} that also sets the privacy option to 'public'.
 It will override default options or options passed as an argument.
 
+== {readonly}
+
+ readonly ssn => my %ssn;
+ readonly fingerprint => my %fingerprint, { %options };
+
+This is an alias to {property} that sets the privacy option to 'public' and
+adds a {set_hook} option that dies if an attempt is made to use the accessor to
+change the property.  It will override default options or options passed as an
+argument.
+
 == {register}
 
  register( bless( $object, $class ) ); # register the object 
@@ -851,6 +909,11 @@ Programmers seeking a more full-featured approach to inside-out objects are
 encouraged to explore [Object::InsideOut].  Other implementations are also
 noted in [Class::InsideOut::Manual::About].
 
+= KNOWN LIMITATIONS
+
+Requires weak reference support (Perl >= 5.6) and Scalar::Util::weaken() to
+avoid memory leaks and to provide thread-safety.
+
 = ROADMAP
 
 Features slated for after the 1.0 release include:
@@ -861,9 +924,7 @@ Features slated for after the 1.0 release include:
 
 = BUGS
 
-Please report bugs or feature requests using the CPAN Request Tracker.
-Bugs can be sent by email to {bug-Class-InsideOut@rt.cpan.org} or
-submitted using the web interface at
+Please report bugs or feature requests using the CPAN Request Tracker:
 [http://rt.cpan.org/Public/Dist/Display.html?Name=Class-InsideOut]
 
 When submitting a bug or request, please include a test-file or a patch to an
@@ -873,42 +934,20 @@ existing test-file that illustrates the bug or desired feature.
 
 David A. Golden (DAGOLDEN)
 
-dagolden@cpan.org
-
-http://dagolden.com/
-
 = COPYRIGHT AND LICENSE
 
-Copyright (c) 2006 by David A. Golden
+Copyright (c) 2006, 2007 by David A. Golden
 
-This program is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself.
+Licensed under the Apache License, Version 2.0 (the "License");
+you may not use this file except in compliance with the License.
+You may obtain a copy of the License at 
+L<http://www.apache.org/licenses/LICENSE-2.0>
 
-The full text of the license can be found in the LICENSE file included with
-this module.
-
-= DISCLAIMER OF WARRANTY
-
-BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
-FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
-OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
-PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
-EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
-WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
-ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
-YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
-NECESSARY SERVICING, REPAIR, OR CORRECTION.
-
-IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
-WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
-REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
-LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
-OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
-THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
-RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
-FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
-SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
-SUCH DAMAGES.
+Unless required by applicable law or agreed to in writing, software
+distributed under the License is distributed on an "AS IS" BASIS,
+WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+See the License for the specific language governing permissions and
+limitations under the License.
 
 =end wikidoc
 
